import os
import decimal
from importlib import import_module
from typing import Any, Optional, List
from flask import Flask, json
from yaml import load, FullLoader
from gammu_smsd_api.blueprints import all_blueprints
from gammu_smsd_api.extensions import gammu_flask
from gammu_smsd_api.config import Config
import gammu_smsd_api as app_root

APP_ROOT_FOLDER = os.path.abspath(os.path.dirname(app_root.__file__))
TEMPLATE_FOLDER = os.path.join(APP_ROOT_FOLDER, 'templates')
STATIC_FOLDER = os.path.join(APP_ROOT_FOLDER, 'static')


def get_config(config_class_string: str, yaml_files: Optional[List[str]] = None) -> Config:
    """Load the Flask config from a class.
    Positional arguments:
    config_class_string -- string representation of a configuration class that will be loaded (e.g.
        'pypi_portal.config.Production').
    yaml_files -- List of YAML files to load. This is for testing, leave None in dev/production.
    Returns:
    A class object to be fed into app.config.from_object().
    """
    config_module, config_class = config_class_string.rsplit('.', 1)
    config_obj = getattr(import_module(config_module), config_class)

    # Load additional configuration settings.
    yaml_files = yaml_files or [f for f in [
        os.path.join('/', 'etc', 'gammu-smsd-api', 'config.yml'),
        os.path.abspath(os.path.join(APP_ROOT_FOLDER, '..', 'config.yml')),
        os.path.join(APP_ROOT_FOLDER, 'config.yml'),
    ] if os.path.exists(f)]
    additional_dict = {}
    for y in yaml_files:
        with open(y, encoding='UTF-8') as f:
            loaded_data = load(f.read(), Loader=FullLoader)
            if isinstance(loaded_data, dict):
                additional_dict.update(loaded_data)
            else:
                raise Exception('Failed to parse configuration {}'.format(y))

    # Merge the rest into the Flask app config.
    for key, value in additional_dict.items():
        setattr(config_obj, key, value)

    return config_obj


def create_app(config_obj: Config) -> Flask:

    """Create an application."""
    app = Flask(__name__, template_folder=TEMPLATE_FOLDER, static_folder=STATIC_FOLDER)
    config_dict = {k: getattr(config_obj, k) for k in dir(config_obj) if not k.startswith('_')}
    app.config.update(config_dict)

    for bp in all_blueprints:
        import_module(bp.import_name)
        app.register_blueprint(bp)

    class FloatJSONEncoder(json.JSONEncoder):
        def default(self, o: Any) -> Any:
            if isinstance(o, decimal.Decimal):
                # Convert decimal instances to float.
                return float(o)
            return super().default(o)

    app.json_encoder = FloatJSONEncoder

    gammu_flask.init_app(app)

    with app.app_context():
        import_module('gammu_smsd_api.middleware')

    return app
